"""
Salesforce OAuth Token Generator
"""

import os
import httpx
from pydantic import BaseModel
from dotenv import load_dotenv
from ..sfauth.sf_token_exchange import SalesforceTokenGenerator

DC_TOKEN_ENDPOINT = '/services/a360/token'


class DCTokenResponse(BaseModel):
    access_token: str
    instance_url: str


class DataCloudTokenGenerator:
    """ Fetches Data Cloud OAuth Access token"""
    def __init__(self, load_env: bool = True, timeout: int = 30):
        """
        Initialize the Data Cloud Token Generator

        Args:
            load_env: Auto load env. variables (default: True)
            timeout: Request timeout in seconds (default: 30)
        """
        if load_env:
            load_dotenv()

        self.dc_token_response = None
        self.sf_inst_url = os.getenv('SALESFORCE_INSTANCE_URL').rstrip('/')
        self.dc_grant_type = os.getenv('DATA_CLOUD_TOKEN_GRANT_TYPE')
        self.subject_token_type = os.getenv('DATA_CLOUD_SUBJECT_TOKEN_TYPE')
        self.timeout = timeout
        self.sf_token_generator = SalesforceTokenGenerator()

        # Initialize HTTP client with common headers
        self.client = httpx.Client(
            timeout=timeout,
            headers={
                'Content-Type': 'application/x-www-form-urlencoded'
            }
        )

    def get_dc_token(self):
        """
        Generates a Salesforce OAuth Token and then exchange that for a DC Access Token
        """
        if self.dc_token_response:
            return self.dc_token_response['access_token']

        sf_token = self.sf_token_generator.get_sf_token()
        url = f"{self.sf_inst_url if self.sf_inst_url else sf_token.instance_url}{DC_TOKEN_ENDPOINT}"

        payload = {
            "grant_type": self.dc_grant_type,
            "subject_token": sf_token.access_token,
            "subject_token_type": self.subject_token_type
        }

        try:
            response = self.client.post(url, data=payload)
            response.raise_for_status()
            json = response.json()
            self.dc_token_response = DCTokenResponse(access_token=json["access_token"],
                                                     instance_url=json["instance_url"])
            return self.dc_token_response

        except httpx.HTTPStatusError as e:
            error_msg = f"HTTP error {e.response.status_code}: {e.response.text}"
            raise httpx.HTTPStatusError(error_msg, request=e.request, response=e.response)
        except httpx.RequestError as e:
            raise httpx.RequestError(f"Network error: {str(e)}")

    def _validate_environment_variables(self):
        """Validate that all required environment variables are set"""
        required_vars = {
            'SALESFORCE_INSTANCE_URL': self.sf_inst_url
        }

        missing_vars = [var for var, value in required_vars.items() if not value]

        if missing_vars:
            raise ValueError(f"Missing required environment variables: {', '.join(missing_vars)}")


"""
JWT Token creator for Salesforce OAuth JWT bearer flow

Requires a connected to be setup in the corresponding org.

Requires a local private key ".key" file to cryptographically sign the token.

Requires the following env. vars:
SALESFORCE_APP_SUBJECT: log-in username
SALESFORCE_APP_PRIVATE_KEY: location of the local private key file
SALESFORCE_APP_CONSUMER_KEY: the consumer key of the connected app
SALESFORCE_ORGANIZATION_ID: 18-char Salesforce org-id
"""

import os
import jwt
import time
import uuid
from pathlib import Path
from dotenv import load_dotenv

TOKEN_EXPIRATION_MINUTES = 5

class SalesforceJWTTokenCreator:
    """creates JWT tokens for Salesforce OAuth JWT bearer flow"""

    def __init__(self, load_env: bool = True):
        """

        :param load_env:
        """
        if load_env:
            load_dotenv()

        self.token = None
        self.app_subject = os.getenv('SALESFORCE_APP_SUBJECT')
        self.organization_id = os.getenv('SALESFORCE_ORGANIZATION_ID')
        self.private_key_path = os.getenv('SALESFORCE_APP_PRIVATE_KEY')
        self.consumer_key = os.getenv('SALESFORCE_APP_CONSUMER_KEY')

        # validate required env. vars
        self._validate_env_vars()

        # load private key
        self.private_key = self._load_private_key()

    def generate_jwt(self,
                     audience: str = "https://login.salesforce.com") -> str:
        """
        generates a signed JWT for the OAuth jwt bearer flow

        :param audience:

        :return:
            str: JWT

        :raises:
            Exception: if there's any error when generating the token
        """
        if self.token and self._token_valid():
            return self.token

        current_time = int(time.time())

        payload = {
            'jti': str(uuid.uuid4()),
            'iss': self.consumer_key,
            'sub': self.app_subject,
            'aud': audience,
            'exp': current_time + (TOKEN_EXPIRATION_MINUTES * 60),
            'iat':current_time
        }

        try:
            token = jwt.encode(
                payload=payload,
                key=self.private_key,
                algorithm='RS356'
            )
            self.token = token
            return token
        except Exception as e:
            raise Exception(f"Exception when generating token: {str(e)}")

    def _validate_env_vars(self):
        requires_vars = {
            'SALESFORCE_APP_SUBJECT': self.app_subject,
            'SALESFORCE_APP_PRIVATE_KEY': self.private_key,
            'SALESFORCE_APP_CONSUMER_KEY': self.consumer_key,
            'SALESFORCE_ORGANIZATION_ID': self.organization_id
        }

        missing_vars = [var for var, value in requires_vars.items() if not value]

        if missing_vars:
            raise ValueError(f"Missing required env. vars: {', '.join(missing_vars)}")

    def _load_private_key(self) -> str:
        try:
            private_key_file = Path(self.private_key_path)

            if not private_key_file.exists():
                raise FileNotFoundError(f"Private key file not found: {self.private_key_path}")

            with open(private_key_file, 'r') as f:
                private_key = f.read()

            if not private_key.strip():
                raise ValueError(f"Private key file is empty: {self.private_key_path}")

            return private_key
        except Exception as e:
            raise Exception(f"Error loading private key: {self.private_key_path}")


    def _token_valid(self) -> bool:
        pass


from .datacloud.dc_token_exchange import DataCloudTokenGenerator
from .sfauth.sf_token_exchange import SalesforceTokenGenerator
from .sfauth.jwt_token_creator import SalesforceJWTTokenCreator

__all__ = [
    "DataCloudTokenGenerator",
    "SalesforceTokenGenerator",
    "SalesforceJWTTokenCreator",
]


"""
Salesforce OAuth Token Generator.

Exchanges a JWT for a Salesforce bearer access token over the JWT bearer flow.

Requires the following env. vars:
SALESFORCE_INSTANCE_URL:
SALESFORCE_TOKEN_GRANT_TYPE:
"""

import os
import logging
import httpx
from dotenv import load_dotenv
from pydantic import BaseModel
from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
    retry_if_exception_type,
    retry_if_exception,
    before_sleep_log,
    after_log
)
from .jwt_token_creator import SalesforceJWTTokenCreator

TOKEN_ENDPOINT = '/services/oauth2/token'
DEFAULT_SF_GRANT_TYPE = 'urn:ietf:params:oauth:grant-type:jwt-bearer'

logger = logging.getLogger(__name__)

class SFTokenResponse(BaseModel):
    access_token: str
    scope: str
    instance_url: str
    id: str
    token_type: str
    api_instance_url: str

def is_retryable_http_error(exception):
    if isinstance(exception, httpx.HTTPStatusError):
        status_code = exception.response.status_code
        return status_code >= 500 or status_code in [408, 429]
    return False

class SalesforceTokenGenerator:
    """
    Fetches SF OAuth Access token over the JWT Bearer flow
    """

    def __init__(self, load_env: bool = True, timeout: int = 30):
        if load_env:
            load_dotenv()

        self.sf_token_response = None
        self.sf_inst_url = os.getenv('SALESFORCE_INSTANCE_URL').rstrip('/')
        self.sf_grant_type = os.getenv('SALESFORCE_TOKEN_GRANT_TYPE', default=DEFAULT_SF_GRANT_TYPE)
        self.timeout = timeout
        self.jwt_creator = SalesforceJWTTokenCreator()

        self.client = httpx.Client(
            timeout=httpx.Timeout(timeout, connect=10.0),
            headers={
                'Content-Type': 'application/x-www-form-urlencoded'
            },
            limits=httpx.Limits(max_keepalive_connections=10, max_connections=20)
        )

    def get_sf_token(self) -> SFTokenResponse:
        """
        Generates a local JWT bearer token and then exchanges that for an access token
        :return:
            SFTokenResponse
        """
        # @todo add expiration validation
        if self.sf_token_response:
            return self.sf_token_response

        self.sf_token_response = self._fetch()
        return self.sf_token_response

    @retry(
        stop=stop_after_attempt(4), # initial attempt + 3 retries
        wait=wait_exponential(multiplier=1, min=1, max=60),
        retry=(
            retry_if_exception_type(httpx.RequestError) |
            retry_if_exception(is_retryable_http_error)
        ),
        before_sleep=before_sleep_log(logger, logging.WARNING),
        after=after_log(logger, logging.DEBUG),
        reraise=True
    )
    def _fetch(self):
        url = f"{self.sf_inst_url}{TOKEN_ENDPOINT}"
        payload = {
            "grant_type": self.sf_grant_type,
            "assertion": self.jwt_creator.generate_jwt()
        }

        logger.debug(f"Attempting to fetch Salesforce Access token: {url}")

        try:
            response = self.client.post(url, data=payload)
            response.raise_for_status()

            token_response = SFTokenResponse(**response.json())
            logger.info("Successfully obtained Salesforce access token")
            return token_response
        except httpx.HTTPStatusError as e:
            status_code = e.response.status_code
            error_txt = e.response.text

            logger.error(f"Http Error {status_code}: {error_txt[:200]}")

            if status_code < 500 and status_code not in [408, 429]:
                raise httpx.HTTPStatusError(
                    f"Non retryable http error {status_code}: {error_txt}",
                    request=e.request,
                    response=e.response
                )
            raise
        except httpx.RequestError as e:
            logger.warning(f"Networking error: {str(e)}")
            raise